# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import time
from tqdm import tqdm


PROMPT_LENGTH = [64, 2048]
MAX_NEW_TOKENS = [64, 2048]
NUM_RUNS = 5
WARMUP_RUNS = 3


def benchmark(model, tokenizer):
    """
    Benchmarkes the throughput of the model. Does some warmup before measuring, to remove compilation time (if
    applicable).
    """
    max_prompt_length = max(PROMPT_LENGTH)
    model_inputs = tokenizer(
        ["foo bar " * 4000], return_tensors="pt", truncation=True, max_length=max_prompt_length
    )

    # sanity check
    if model_inputs.input_ids.shape[1] != max_prompt_length:
        raise ValueError(
            f"Benchmark error: Model input length is {model_inputs.input_ids.shape[1]}, but expected to be "
            f"{max_prompt_length}."
        )

    # benchmark
    results = {}
    for prompt_length in PROMPT_LENGTH:
        for max_new_tokens in MAX_NEW_TOKENS:
            print(f"\nBenchmarking with prompt_length={prompt_length} and max_new_tokens={max_new_tokens}.")
            run_name = f"prompt_length={prompt_length}, max_new_tokens={max_new_tokens}"

            input_ids = model_inputs.input_ids[:, :prompt_length].to(model.device)
            for _ in tqdm(range(WARMUP_RUNS), desc="Warming up"):
                model.generate(
                    input_ids, do_sample=False, max_new_tokens=max_new_tokens, min_new_tokens=max_new_tokens
                )

            tokens_per_second = []
            for _ in tqdm(range(NUM_RUNS), desc="Benchmarking"):
                start = time()
                gen_out = model.generate(
                    input_ids, do_sample=False, max_new_tokens=max_new_tokens, min_new_tokens=max_new_tokens
                )
                end = time()
                if gen_out.shape[1] != prompt_length + max_new_tokens:
                    raise ValueError(
                        f"Benchmark error: Generated output length is {gen_out.shape[1]}, but expected to be "
                        f"{prompt_length + max_new_tokens}."
                    )
                tokens_per_second.append((max_new_tokens) / (end - start))
            results[run_name] = sum(tokens_per_second)/len(tokens_per_second)
            print(f"{run_name:40s}: {results[run_name]:2f} tokens per second.\n")

    # print results
    print("\n\nResults:")
    for run_name, throughput in results.items():
        print(f"{run_name:40s}: {throughput:2f} tokens per second.")



# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys

import torch
from transformers import AutoTokenizer, TextStreamer, set_seed
from transformers.utils import logging

from local_gemma import LocalGemma2ForCausalLM
from .utils.benchmark import benchmark
from .utils.config import infer_device, infer_dtype, get_prompt, get_generation_kwargs, infer_memory_requirements


MODEL_NAMES = {
    "9b": "google/gemma-2-9b-it",
    "27b": "google/gemma-2-27b-it",
}

parser = argparse.ArgumentParser(description="Local Gemma")

# Prompt argument
parser.add_argument(
    "prompt",
    type=str,
    nargs="*",
    help=(
        "Prompt to the model. For an interactive session, leave this field empty."
    ),
)
# Other control arguments
parser.add_argument(
    "--model",
    type=str,
    default="9b",
    help=(
        "Size of Gemma 2 instruct model to be used in the application ('9b' or '27'b) or, alternatively, a Hugging "
        "Face repo. Defaults to '9b'."
    ),
)
parser.add_argument(
    "--token",
    type=str,
    help="Authentication token for the model. Required to download the model into a local cache.",
)
parser.add_argument(
    "--preset",
    type=str,
    choices=["auto", "exact", "speed", "memory", "memory_extreme"],
    default="auto",
    help=(
        "Sets the optimization strategy for the local model deployment. Defaults to 'auto', which selects the best "
        "strategy for your device. 'exact' maximises accuracy, 'speed' maximises generation speed, 'memory' reduces "
        "memory requirements through quantization, and 'memory_extreme' minimises memory requirements."
    ),
)
parser.add_argument(
    "--mode",
    type=str,
    choices=["chat", "factual", "creative"],
    default="chat",
    help=(
        "Sets the mode of operation of the model. 'chat' is optimized for general conversation, 'factual' is designed "
        "to minimize hallucinations, and 'creative' is optimized for creative writing. Note that 'factual' and "
        "'creative' prepend text to your prompt."
    ),
)
parser.add_argument(
    "--max_new_tokens",
    type=int,
    help=(
        "Maximum number of tokens to be used in each generation round. By default it relies on the model to emit an "
        "EOS token, and generates up to the pretrained length."
    ),
)
parser.add_argument(
    "--device",
    type=str,
    help="Forces a specific device to be used. By default uses cuda > mps > cpu, depending on availability.",
)
parser.add_argument(
    "--dtype",
    type=str,
    help="The dtype in which computations are performed. Defaults to the dtype set by --preset",
)
parser.add_argument(
    "--silent",
    action="store_true",
    help="Does NOT print any output except for the model outputs.",
)
# Debugging arguments
parser.add_argument(
    "--seed",
    type=int,
    help="Seed for text generation. Optional, use for reproducibility.",
)
parser.add_argument(
    "--benchmark",
    action="store_true",
    help="Runs a throughput benchmark on your device.",
)


def print_help(is_instruction_tuned: bool = True):
    if is_instruction_tuned:
        print("\nYou can now interact with the model through a conversation. A few tips:")
        print("- Initialize the program with '--silent' to hide all non-model messages")
        print("- Input '!exit' to leave the program")
        print("- Input '!new session' to reset the conversation")
        print("- Input '!help' to print this message again")
    else:
        print("\nYou can now pass a prompt to the base model to generate a single response.")
        print("Tip: for multi-turn conversation, use an instruction tuned model, such as `google/gemma-2-9b-it`.")
    print("")


def main():
    args = parser.parse_args()

    stdout_received = not sys.stdin.isatty()
    if stdout_received:
        input_data = sys.stdin.read()
        args.prompt = args.prompt + ["\n"] + [input_data]

    device = infer_device(args.device)
    dtype = infer_dtype(device, args.dtype)
    generation_kwargs = get_generation_kwargs(args.mode)
    base_prompt = get_prompt(args.mode)
    model_name = MODEL_NAMES.get(args.model) or args.model

    if args.preset == "auto":
        args.preset = infer_memory_requirements(model_name, device, trust_remote_code=False, token=args.token)

    # TODO(joao) : assisted generation
    # # Triggers assisted generation on CUDA or MPS devices, assuming the default model is used. Assisted generation is
    # # not beneficial on most CPU settings.
    # if  args.model_name is None and ("cuda" in device or device.isdigit() or "mps" in device):
    #     assistant_model_name = (
    #         ASSISTANT_MODEL_NAME if args.optimization == "quality" else QUANTIZED_ASSISTANT_MODEL_NAME
    #     )
    # else:
    #     assistant_model_name = None

    if not args.silent:
        print("\nLoading model with the following characteristics:")
        print("- Model name:", model_name)
        # print("- Assistant model name:", assistant_model_name)
        print("- Device:", device)
        print("- Data type:", str(dtype))
        print("- Optimization preset:", args.preset)
        print("- Generation arguments:", str(generation_kwargs))
        print("- Base prompt:", repr(base_prompt) if len(base_prompt) > 0 else "None")
        print("")
    else:
        logging.disable_progress_bar()

    tokenizer = AutoTokenizer.from_pretrained(model_name, token=args.token)
    model = LocalGemma2ForCausalLM.from_pretrained(
        model_name, preset=args.preset, token=args.token, torch_dtype=dtype, device=device
    )
    # TODO(joao): this if shouldn't be needed, fix in transformers
    model._supports_cache_class = True

    # if assistant_model_name is not None:
    #     assistant_model = LocalGemma2ForCausalLM.from_pretrained(
    #         assistant_model_name, preset=args.preset, token=args.token, torch_dtype=dtype, device=device)
    # else:
        # assistant_model = None
    assistant_model = None

    if args.benchmark:
        benchmark(model, tokenizer)
    else:
        if args.seed is not None:
            set_seed(args.seed)

        has_starting_prompt = len(args.prompt) > 0
        is_instruction_tuned = tokenizer.chat_template is not None

        if device == "mps" and args.preset == "auto" and args.max_new_tokens is None:
            print("Setting max new tokens to 1024 for faster mps generation. To bypass this limit, set `--max_new_tokens=2048`.")
            args.max_new_tokens = 1024

        if not args.silent and not has_starting_prompt:
            print_help(is_instruction_tuned=is_instruction_tuned)

        streamer = TextStreamer(tokenizer, skip_prompt=True, **{"skip_special_tokens": True})
        cache = None
        chat_history = []
        while True:
            # Get input to the model
            if has_starting_prompt:
                user_input = " ".join(args.prompt)
            else:
                user_input = input(">>> ")

            # Handle special commands
            if user_input in ["!exit", "quit", "quit()"]:
                break
            elif user_input == "!new session":
                chat_history = []
                if hasattr(cache, "reset"):
                    cache.reset()
                else:
                    cache = None
            elif user_input == "!help":
                print_help()

            # Generate text
            else:
                # Inject the base prompt if the chat history is empty
                if len(chat_history) == 0:
                    user_input = base_prompt + user_input

                chat_history += [{"role": "user", "content": user_input},]
                if is_instruction_tuned:
                    tokenized_chat = tokenizer.apply_chat_template(
                        chat_history, tokenize=True, add_generation_prompt=True, return_tensors="pt"
                    )
                else:
                    tokenized_chat = tokenizer(user_input, return_tensors="pt").input_ids

                tokenized_chat = tokenized_chat.to(device)
                generation_kwargs.update(
                    {
                        "streamer": streamer,
                        "assistant_model": assistant_model,
                        "return_dict_in_generate": True,
                        "past_key_values": cache,
                    }
                )
                # TODO(joao): this if shouldn't be needed, fix in transformers
                if cache is not None:
                    generation_kwargs["cache_implementation"] = None

                if args.max_new_tokens is not None:
                    generation_kwargs["max_new_tokens"] = args.max_new_tokens
                    input_ids_len = tokenized_chat.shape[-1]
                    max_cache_len = args.max_new_tokens + input_ids_len
                    if cache is not None and cache.max_cache_len < max_cache_len:
                        # reset the cache
                        generation_kwargs.pop("past_key_values")
                        generation_kwargs["cache_implementation"] = "hybrid"
                else:
                    generation_kwargs["max_length"] = model.config.max_position_embeddings

                if device == "mps":
                    generation_kwargs["attention_mask"] = torch.ones_like(tokenized_chat)

                gen_out = model.generate(input_ids=tokenized_chat, **generation_kwargs)

                # Store the cache for the next generation round; Pull the model output into the chat history.
                cache = gen_out.past_key_values
                model_tokens = gen_out.sequences[0, tokenized_chat.shape[1]:]
                model_output_text = tokenizer.decode(model_tokens, skip_special_tokens=True)
                chat_history += [{"role": "assistant", "content": model_output_text},]

                if is_instruction_tuned:
                    # Sanity check: EOS was removed, ends in "<end_of_turn>\n"
                    tokenized_chat = tokenizer.apply_chat_template(
                        chat_history, tokenize=True, add_generation_prompt=False, return_tensors="pt"
                    ).tolist()[0]
                    assert tokenized_chat[0] == 2
                    assert tokenized_chat[-1] == 108
                    assert tokenized_chat[-2] == 107

            if has_starting_prompt or not is_instruction_tuned:
                break

if __name__ == '__main__':
    main()



# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import psutil
import torch

from typing import Dict, Optional

from accelerate import init_empty_weights
from transformers import Gemma2ForCausalLM, Gemma2Config
from transformers.utils import is_torch_bf16_available_on_device
from accelerate.utils import calculate_maximum_sizes


DTYPE_MODIFIER = {"exact": 2, "memory": 8, "memory_extreme": 8}
DTYPE_MAP = {"float32": torch.float32, "float16": torch.float16, "bfloat16": torch.bfloat16}


def infer_device(device: Optional[str] = None) -> str:
    """
    Infers basic devices available on the system. Prioritizes the most performant device.
    """
    if device is not None:
        return device
    elif torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    return "cpu"


def infer_dtype(device: str, dtype_str: Optional[str] = None) -> torch.dtype:
    if dtype_str is None:
        if is_torch_bf16_available_on_device(device):
            return torch.bfloat16
        else:
            return torch.float16
    dtype = DTYPE_MAP.get(dtype_str, None)
    if dtype is None:
        raise ValueError(f"Unknown dtype: {dtype_str}. Must be one of {DTYPE_MAP.keys()}")
    return dtype


def get_prompt(mode: str) -> str:
    if mode == "chat":
        return ""
    elif mode == "factual":
        return "Please reply to the following requests with short and factual answers.\n\n"
    elif mode == "creative":
        return (
            "Write a response that appropriately completes the request. Be descriptive, fluid, and follow the context "
            "provided.\n\n"
        )
    else:
        raise ValueError(f"Unknown mode: {mode}")


def get_generation_kwargs(mode: str) -> Dict:
    generation_kwargs = {"do_sample": True}
    if mode == "chat":
        generation_kwargs["temperature"] = 0.7
    elif mode == "factual":
        generation_kwargs["temperature"] = 0.3
        generation_kwargs["repetition_penalty"] = 1.2
    elif mode == "creative":
        generation_kwargs["min_p"] = 0.08
        generation_kwargs["temperature"] = 1.5
    else:
        raise ValueError(f"Unknown mode: {mode}")
    return generation_kwargs


def infer_memory_requirements(model_name, device=None, token=None, trust_remote_code=False) -> str:
    config = Gemma2Config.from_pretrained(model_name, token=token, trust_remote_code=trust_remote_code)

    with init_empty_weights():
        model = Gemma2ForCausalLM(config)

    total_size, _ = calculate_maximum_sizes(model)
    device = infer_device(device)

    if device == "cuda":
        total_memory = torch.cuda.get_device_properties(device).total_memory
    else:
        total_memory = psutil.virtual_memory().total

    for preset in DTYPE_MODIFIER.keys():
        dtype_total_size = total_size / DTYPE_MODIFIER[preset]
        inference_requirements = 1.2 * dtype_total_size

        if inference_requirements < total_memory:
            return preset

    # if the model does not fit fully in the device, return the last preset ('memory_extreme') which will automatically
    # enable CPU offloading so that we can fit any device
    return "memory_extreme"



# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

from typing import Optional, Union, Dict
import logging

import torch
from transformers import QuantoConfig, is_bitsandbytes_available, BitsAndBytesConfig
from transformers.utils import is_quanto_available, is_torch_sdpa_available, is_accelerate_available
from transformers.models.gemma2 import Gemma2ForCausalLM, Gemma2Config
from .utils.config import infer_device, infer_dtype, infer_memory_requirements


logger = logging.getLogger(__name__)

EXACT = {
    "attn_implementation": "eager",
}

MEMORY = {
    "attn_implementation": "eager",
    "quantization_config": {
        "weights": "int4"
    }
}

MEMORY_EXTREME = {
    "attn_implementation": "eager",
    "device_map": "auto",
    "quantization_config": {
        "weights": "int4"
    }
}


PRESET_MAPPING = {
    "auto": None,
    "exact": EXACT,
    "memory": MEMORY,
    "memory_extreme": MEMORY_EXTREME,
}

class LocalGemma2ForCausalLM(Gemma2ForCausalLM):
    @staticmethod
    def get_preset_kwargs(pretrained_model_name_or_path: str, preset: str, device: str, trust_remote_code: bool = False, token: str = None) -> Dict:
        if preset not in PRESET_MAPPING:
            raise ValueError(f"Got invalid `preset` {preset}. Ensure `preset` is one of: {list(PRESET_MAPPING.keys())}")

        if preset == "auto":
            preset = infer_memory_requirements(pretrained_model_name_or_path, device, trust_remote_code=trust_remote_code, token=token)
            logger.info(f"Detected device {device} and defaulting to {preset} preset.")

        preset_kwargs = PRESET_MAPPING[preset]

        if preset in ["memory", "memory_extreme"]:
            if not is_torch_sdpa_available():
                logger.warning_once(
                    "Detected PyTorch version <2.1.1. For faster inference through SDPA attention, install PyTorch "
                    "v2.1.1 or later through the official instructions: https://pytorch.org/"
                )
                preset_kwargs["attn_implementation"] = "eager"
            if device == "cuda" and not is_bitsandbytes_available():
                raise ImportError(
                    f"The {preset} preset on CUDA requires the `bitsandbytes` package. Please install bitsandbytes through: "
                    "`pip install --upgrade bitsandbytes`."
                )
            elif device != "cuda" and not is_quanto_available():
                raise ImportError(
                    f"The {preset} preset on {device} requires the `quanto` package. Please install quanto through: "
                    "`pip install --upgrade quanto`."
                )

        if preset == "memory_extreme":
            if not is_accelerate_available():
                raise ImportError(
                    f"The `memory_extreme` preset requires the `accelerate` package. Please install accelerate through: "
                    "`pip install --upgrade accelerate`."
                )

        return preset_kwargs

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: Optional[Union[str, os.PathLike]],
        preset: Optional[str] = "auto",
        *model_args,
        config: Optional[Union[Gemma2Config, str, os.PathLike]] = None,
        cache_dir: Optional[Union[str, os.PathLike]] = None,
        ignore_mismatched_sizes: bool = False,
        force_download: bool = False,
        local_files_only: bool = False,
        token: Optional[Union[str, bool]] = None,
        revision: str = "main",
        use_safetensors: bool = None,
        **kwargs,
    ) -> Gemma2ForCausalLM:
        device = infer_device(kwargs.pop("device", None))
        preset_kwargs = cls.get_preset_kwargs(
            pretrained_model_name_or_path,
            preset,
            device=device,
            trust_remote_code=kwargs.get("trust_remote_code"),
            token=kwargs.get("token"),
        )

        preset_kwargs["low_cpu_mem_usage"] = True
        torch_dtype = kwargs.pop("torch_dtype", None)
        if torch_dtype is None:
            torch_dtype = infer_dtype(device)
            if torch_dtype == torch.float16:
                extra_message = ' and weights' if preset not in ['memory', 'memory_extreme'] else ''
                logger.warning(
                    f"Defaulting to float16 precision for the computations{extra_message}. "
                    f"This can cause instabilities in generation for larger models, e.g. the 27b checkpoints."
                )
        preset_kwargs["torch_dtype"] = torch_dtype

        quantization_config = kwargs.pop("quantization_config", None)
        if quantization_config is not None:
            preset_kwargs["quantization_config"] = quantization_config
        elif preset_kwargs.get("quantization_config"):
            if device == "cuda":
                preset_kwargs["quantization_config"] = BitsAndBytesConfig(
                    load_in_4bit=True,
                    bnb_4bit_compute_dtype=preset_kwargs["torch_dtype"],
                )
            else:
                preset_kwargs["quantization_config"] = QuantoConfig(
                    weights=preset_kwargs["quantization_config"]["weights"]
                )

        # give preference to kwargs passed by the user
        kwargs_copy = kwargs.copy()
        if kwargs is not None:
            for key in kwargs_copy:
                if key in preset_kwargs:
                    preset_kwargs[key] = kwargs.pop(key)

        model = super().from_pretrained(
            pretrained_model_name_or_path,
            *model_args,
            config=config,
            cache_dir=cache_dir,
            ignore_mismatched_sizes=ignore_mismatched_sizes,
            force_download=force_download,
            local_files_only=local_files_only,
            token=token,
            revision=revision,
            use_safetensors=use_safetensors,
            **preset_kwargs,
            **kwargs,
        )

        if device not in str(model.device) and preset_kwargs.get("device_map", None) is None:
            # for consistent behaviour with bitsandbytes, we move the model to the device always
            model.to(device, dtype=preset_kwargs["torch_dtype"])

        return model



# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools


DEPS = [
    "setuptools",
    "torch>=2.1.1",
    "accelerate",
    "transformers>=4.42.3",
]

EXTRA_CUDA_DEPS = ["bitsandbytes"]
EXTRA_MPS_DEPS = ["quanto>=0.2.0"]

setuptools.setup(
    name='local_gemma',
    version='0.1.0',
    author="The Hugging Face team",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': ['local-gemma=local_gemma.cli:main']
        },
    install_requires=DEPS,
    extras_require={
        "cuda": EXTRA_CUDA_DEPS,
        "mps": EXTRA_MPS_DEPS,
        "cpu": EXTRA_MPS_DEPS,
    },
)
